(ns wanijo.instance.domain
  (:require [clojure.spec.alpha :as spec]
            [wanijo.framework.neo4j :as neo4j]
            [clojure.pprint :as pprint]))

(spec/def ::created-at ::neo4j/date-str)
(spec/def ::updated-at ::neo4j/date-str)
(spec/def ::name (spec/and (complement empty?) string?))

(neo4j/defquery findy-by-schema
  "MATCH (i:instance)-[:of]->(s:schema)
   WHERE s.uuid = {uuid}
   RETURN i
   ORDER BY i.updated_at DESC")

(defn find-by-schema! [schema-uuid]
  (map :i
       (neo4j/exec-query!
        findy-by-schema
        {:uuid schema-uuid})))

(neo4j/defquery create-instance
  "MATCH (s:schema {uuid:{schema_uuid}})
   CREATE (i:instance {uuid:{uuid}})-[:of]->(s)
      SET i.name = {name},
          i.created_at = {created_at},
          i.updated_at = {created_at}")

(neo4j/defquery create-property
  "MATCH (i:instance {uuid:{uuid}}),
         (a:attribute {uuid:{attr_uuid}})
   CREATE (p:property {uuid:{prop_uuid}})-[:of]->(i),
          (p)-[:of]->(a)
     SET p.value = {value},
         p.created_at = {created_at},
         p.updated_at = {updated_at}")

(defn create! [schema-uuid instance]
  (let [instance-uuid (neo4j/uuid)
        now (neo4j/now-str)
        instance-tuple [create-instance
                        {:schema_uuid schema-uuid
                         :name (:name instance)
                         :uuid instance-uuid
                         :created_at now}]
        prop-tuples (for [{:keys [attribute value]} (:properties instance)]
                      [create-property
                       {:uuid instance-uuid
                        :attr_uuid (:uuid attribute)
                        :prop_uuid (neo4j/uuid)
                        :value value
                        :created_at now}])]
    (apply neo4j/exec-queries!
           (concat [instance-tuple]
                   prop-tuples))))

(neo4j/defquery find-by-uuid
  "MATCH (i:instance {uuid:{uuid}})
         -[:of]->(s:schema)
   RETURN i, s")

(defn find-by-uuid! [uuid]
  (->> (neo4j/exec-query! find-by-uuid
                          {:uuid uuid})
       (map #(assoc (:i %)
                    :schema
                    (:s %)))
       first))

(neo4j/defquery find-properties
  "MATCH (i:instance {uuid:{uuid}}),
         (p:property)-[:of]->(i),
         (p)-[:of]->(a:attribute)
   RETURN p, a
    ORDER BY a.name")

(defn find-properties! [uuid]
  (map #(assoc (:p %)
               :attribute
               (:a %))
       (neo4j/exec-query! find-properties
                          {:uuid uuid})))

(neo4j/defquery edit-instance
  "MATCH (i:instance {uuid:{uuid}})
     SET i.name = {name},
         i.updated_at = {updated_at}")

(neo4j/defquery edit-property
  "MATCH (i:instance {uuid:{instance_uuid}}),
         (a:attribute {uuid:{attribute_uuid}})
   MERGE (p:property {uuid:{uuid}})-[:of]->(i)
   MERGE (p)-[:of]->(a)
   ON CREATE SET p.created_at = {now},
                 p.updated_at = {now},
                 p.value = {value}
   ON MATCH SET p.updated_at = {now},
                p.value = {value}")

(defn edit! [instance]
  (let [prop-tuples (map #(vector edit-property
                                  {:uuid (:uuid % (neo4j/uuid))
                                   :now (neo4j/now-str)
                                   :value (:value %)
                                   :instance_uuid (:uuid instance)
                                   :attribute_uuid (-> % :attribute :uuid)})
                         (:properties instance))]
    (apply neo4j/exec-queries!
           (concat [[edit-instance
                     {:uuid (:uuid instance)
                      :name (:name instance)
                      :updated_at (neo4j/now-str)}]]
                   prop-tuples))))

(neo4j/defquery delete
  "MATCH (i:instance {uuid:{uuid}}),
         (i)-[ic:of]->(s:schema)
   OPTIONAL MATCH
         (p:property)-[pc:of]->(i),
         (p)-[pac:of]->(a:attribute)
   DELETE pac, pc, ic, p, i")

(defn delete! [uuid]
  (neo4j/exec-query! delete {:uuid uuid}))

(neo4j/defquery create-link
  "MATCH (i:instance {uuid:{from}}),
         (u:user {uuid:{by}}),
         (t:instance)
   WHERE t.uuid IN {to}
   CREATE (l:link {uuid:{uuid}})-[:created_by]->(u)
      SET l.created_at = {created_at},
          l.name = {name}
   CREATE (i)<-[:link_from]-(l)-[:link_to]->(t)")

(defn create-link! [link]
  (neo4j/exec-query! create-link
                     (assoc link
                            :uuid (neo4j/uuid)
                            :created_at (neo4j/now-str))))
